from typing import Union, List

import numpy as np
from sympy import sympify, expand, Expr, Basic
from sympy.core import Mul, Add, Pow, Symbol
from sympy import symbols as sp_symbols
import re

from SRToolkit.utils.symbol_library import SymbolLibrary
from SRToolkit.utils.expression_tree import Node
from SRToolkit.utils.expression_tree import is_float

def simplify(expr: Union[List[str], Node], symbol_library: SymbolLibrary=SymbolLibrary.default_symbols()) -> Union[List[str], Node]:
    """
    Simplifies a mathematical expression by:
        1. making use of sympy's simplification functions
        2. simplifying constants, e.g. C*C + C -> C

    Examples:
        >>> expr = ["C", "+", "C" "*", "C", "+", "X_0", "*", "X_1", "/", "X_0"]
        >>> print("".join(simplify(expr)))
        C+X_1
    
    Args:
        expr: The expression given as a list of tokens in the infix notation or as an instance of SRToolkit.utils.expression_tree.Node
        symbol_library: The symbol library to use. Defaults to SymbolLibrary.default_symbols().

    Raises:
        Exception: If problems occur during simplification or if the expression contains invalid symbols.

    Returns:
        The simplified expression
    """
    is_tree = False
    if isinstance(expr, Node):
        expr = expr.to_list(symbol_library=symbol_library, notation="infix")
        is_tree = True

    variables = symbol_library.get_symbols_of_type("var")

    # We expect only one symbol for constants
    if len(symbol_library.get_symbols_of_type("const")) > 0:
        constant = symbol_library.get_symbols_of_type("const")[0]
    else:
        # In this case constants shouldn't be problematic as they are not in the SymbolLibrary
        # Just in case and to not change other functions, I changed it to __C__.
        constant = "__C__"

    expr = _simplify_expression("".join(expr), constant, variables)
    expr = sympify(_denumerate_constants(str(expr), constant), evaluate=False)
    expr = _sympy_to_sr(expr)
    if not _check_tree(expr, symbol_library):
        raise Exception("Simplified expression contains invalid symbols. Possibly skip its simplification or add symbols to the SymbolLibrary.")

    if is_tree:
        return expr
    else:
        return expr.to_list(symbol_library=symbol_library, notation="infix")


def _check_tree(expr: Node, symbol_library: SymbolLibrary) -> bool:
    """
    Checks if the expression tree contains only valid symbols from the symbol library.

    Args:
        expr: The expression tree to check.
        symbol_library: The symbol library to use.

    Returns:
        True if the expression tree contains only valid symbols from the symbol library, False otherwise.
    """
    if expr.symbol not in symbol_library.symbols and not is_float(expr.symbol):
        return False
    if isinstance(expr.left, Node) and not _check_tree(expr.left, symbol_library):
        return False
    if isinstance(expr.right, Node) and not _check_tree(expr.right, symbol_library):
        return False

    return True


def _sympy_to_number(expr):
    """
    Extracts the number contained in the Sympy node 
    """
    evaluated = float(expr.evalf())
    return int(evaluated) if evaluated.is_integer() else evaluated


def _sympy_to_sr(expr: Union[Expr, Basic]) -> Node:
    """
    Converts a Sympy expression into an instance of SRtoolkit.expression_tree.Node, explicitly handling left-associative division.

    Args:
        expr: The Sympy expression.

    Returns:
        An instance of SRtoolkit.expression_tree.Node that corresponds to the expression given as the input.
    """
    if expr.is_Number:
        return Node(str(_sympy_to_number(expr)))
    
    if expr.is_Symbol:
        return Node(str(expr))

    if expr.is_Function:
        func_name = expr.func.__name__
        arg = _sympy_to_sr(expr.args[0])
        return Node(func_name, left=arg)

    if isinstance(expr, Add):
        args = expr.as_ordered_terms()
        # Detect subtraction
        if len(args) == 2 and args[1].is_Mul and args[1].args[0] == -1:
            return Node('-', _sympy_to_sr(-args[1]), _sympy_to_sr(args[0]))
        # Handle regular addition
        root = Node('+', _sympy_to_sr(args[1]), _sympy_to_sr(args[0]))
        for term in args[2:]:
            root = Node('+', _sympy_to_sr(term), root)
        return root

    if isinstance(expr, Mul):
        # Process factors explicitly, ensuring left-to-right associativity
        factors = list(expr.args)
        root = _sympy_to_sr(factors[0])  # Start with the first factor
        for factor in factors[1:]:
            if factor.is_Pow and factor.args[1] == -1:  # Division
                divisor = _sympy_to_sr(factor.args[0])
                root = Node('/', divisor, root)  # Left-to-right division
            else:  # Multiplication
                multiplicand = _sympy_to_sr(factor)
                root = Node('*', multiplicand, root)
        return root

    if isinstance(expr, Pow):
        base, exp = expr.args
        return Node('^', _sympy_to_sr(exp), _sympy_to_sr(base))

    if expr.is_Rational and expr.q != 1:
        return Node('/', _sympy_to_sr(expr.q), _sympy_to_sr(expr.p))

    raise ValueError(f"{expr}")


def _simplify_constants(eq, c, var):
    """ Simplifies the constants in a Sympy expression. output[2][0][1] is the simplified expression.

    Args:
        eq: The Sympy expression.
        c: The constant symbol.
        var: List of symbols representing variables.
    
    Returns:
        - bool: True if the expression contains a variable.
        - bool: True if the expression contains the constant.
        - list: List of tuples containing the original and simplified expressions
    """
    if len(eq.args) == 0:
        if eq in var:
            return True, False, [(eq, eq)]
        elif eq in eq.free_symbols:
            return False, True, [(eq, c)]
        else:
            return False, False, [(eq, eq)]
    else:
        has_var, has_c, subs = [], [], []
        for a in eq.args:
            a_rec = _simplify_constants (a, c, var)
            has_var += [a_rec[0]]; has_c += [a_rec[1]]; subs += [a_rec[2]]
        if sum(has_var) == 0 and True in has_c:
            return False, True, [(eq, c)]
        else:          
            args = []
            if isinstance(eq, (Add, Mul, Pow)):
                has_free_c = False
                if True in [has_c[i] and not has_var[i] for i in range(len(has_c))]:
                    has_free_c = True
                    
                for i in range(len(has_var)):
                    if has_var[i] or (not has_free_c and not has_c[i]):
                        if len(subs[i]) > 0:
                            args += [eq.args[i].subs(subs[i])]
                        else:
                            args += [eq.args[i]]
                if has_free_c:
                    args += [c]
                
            else:
                for i in range(len(has_var)):
                    if len(subs[i]) > 0:
                        args += [eq.args[i].subs(subs[i])]
                    else:
                        args += [eq.args[i]]
            return True in has_var, True in has_c, [(eq, eq.func(*args))]


def _enumerate_constants(expr, constant):
    """ Enumerates the constants in a Sympy expression. 

    Example: C*x**2 + C*x + C -> C0*x**2 + C1*x + C2

    Input:
        expr - Sympy expression
        constant - constant symbol

    Returns:
        Sympy expression with enumerated constants
        list of enumerated constants"""
        
    char_list = np.array(list(str(expr)), dtype='<U16')
    constind = np.where(char_list == constant)[0]
    """ Rename all constants: c -> cn, where n is the index of the associated term"""
    constants = [constant+str(i) for i in range(len(constind))]
    char_list[constind] = constants
    return sympify("".join(char_list)), tuple(constants)


def _denumerate_constants(expr, constant):
    """ Removes the enumeration of constants in a Sympy expression.

    Args:
        expr: Sympy expression
        constant: constant symbol

    Returns:
        Sympy expression with denumerated constants
    """
    return re.sub(f'{constant}\\d', constant, expr)


def _simplify_expression (expr_str, constant, variables):
    """Simplifies a mathematical expression.

    Args:
        expr_str: String representing the expression.
        constant: The character representing numerical constants.
        variables: List of characters representing variables.

    Returns:
        expr: Sympy expression object in canonical form.
        symbols_params: Tuple of enumerated constants.
    """
    x = [sp_symbols(s.strip("'")) for s in variables]
    c = sp_symbols(constant)

    expr, _ = _enumerate_constants(expr_str, constant)
    expr = _simplify_constants(expr, c, x)[2][0][1]
    expr, _ = _enumerate_constants(expr, constant)
    expr = expand(expr)
    expr = _simplify_constants(expr, c, x)[2][0][1]
    # expr, _ = _enumerate_constants(expr, constant)
    # expr = _simplify_constants(expr, c, x)[2][0][1]
    return expr


if __name__ == "__main__":
    # Should simplify
    expr = ["C", "+", "C" "*", "C", "+", "X_0", "*", "X_1", "/", "X_0"]
    print(simplify(expr))

    # Should raise an exception
    expr = ["X_0", "*", "X_0", "^2"]
    sl = SymbolLibrary.from_symbol_list(["+", "*", "-"], 1)
    print(simplify(expr, symbol_library=sl))


"""
This module contains the SymbolLibrary class, which is used for managing symbols and their properties.
"""
import copy
from typing import List, Dict


class SymbolLibrary:
    def __init__(self):
        """
        Initializes an instance of the SymbolLibrary class. This class is used for managing symbols and their
        properties for other functionality in this package.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x", "x")
            >>> library.get_type("x")
            'var'
            >>> library.get_precedence("x")
            0
            >>> library.get_np_fn("x")
            'x'
            >>> library.remove_symbol("x")
            >>> library = SymbolLibrary.default_symbols()

        Attributes:
            symbols : dict
                A dictionary mapping symbols to their properties (type, precedence, numpy function).

        Methods:
            add_symbol(symbol, symbol_type, precedence, np_fn):
                Adds a symbol to the library.
            remove_symbol(symbol):
                Removes a symbol from the library.
            get_type(symbol):
                Retrieves the type of a symbol from the library.
            get_precedence(symbol):
                Returns the precedence of the given symbol.
            get_np_fn(symbol):
                Returns the numpy function corresponding to the given symbol.
            default_symbols():
                Returns a SymbolLibrary with the default symbols.
        """
        self.symbols = dict()
        self.num_variables = 0

    def __len__(self) -> int:
        """
        Returns the number of symbols currently stored in the SymbolLibrary.

        Examples:
             >>> library = SymbolLibrary.default_symbols(5)
             >>> len(library)
             34
             >>> library.add_symbol("a", "lit", 5, "a", "a")
             >>> len(library)
             35

        Returns
            Number of symbols currently stored in the SymbolLibrary.
        """
        return len(self.symbols)

    def __str__(self) -> str:
        r"""
        Returns a string representation of the SymbolLibrary instance.

        This method provides a comma-separated string of all the symbol keys
        currently stored in the SymbolLibrary.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x", "x")
            >>> str(library)
            'x'
            >>> library.add_symbol("sin", "fn", 5, "{} = np.sin({})", r"\sin {}")
            >>> str(library)
            'x, sin'

        Returns:
            A string containing all symbols in the library, separated by commas.
        """
        return ", ".join(self.symbols.keys())

    def __copy__(self) -> "SymbolLibrary":
        r"""
        Creates a copy of the SymbolLibrary instance.

        Examples:
            >>> old_symbols = SymbolLibrary()
            >>> old_symbols.add_symbol("x", "var", 0, "x", "x")
            >>> print(old_symbols)
            x
            >>> new_symbols = copy.copy(old_symbols)
            >>> new_symbols.add_symbol("sin", "fn", 5, "{} = np.sin({})", r"\sin {}")
            >>> print(old_symbols)
            x
            >>> print(new_symbols)
            x, sin

        Returns:
            A copy of the SymbolLibrary instance.
        """
        sl = SymbolLibrary()
        sl.symbols = copy.deepcopy(self.symbols)
        return sl

    def add_symbol(self, symbol: str, symbol_type: str, precedence: int, np_fn: str, latex_str: str = None):
        r"""
        Adds a symbol to the library. A symbol should have a type, precedence, a numpy function, and a LaTeX template associated with it.
        Type "op" should be used for symbols operating on two operands, "fn" for symbols operating on one operand,
        "lit" for constants with a known value (such as pi or e), "const" for constants/parameters without a value that
        need to be optimized, and "var" for variables whose values are provided as input data.

        We recommend you use a single token of "const" type as using multiple might lead to more work, errors, and less
        readability.

        If the argument 'latex_str' is ommited, a default LaTeX template will be generated for the symbol. In case of symbol 'symb', the default template
        will be '{} \text{symb} {}' for an operator,'\text{symb} {}' for a function, and '\text{symb}' otherwise.

        For example, look at the default_symbols function for the SymbolLibrary class.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.add_symbol("sin", "fn", 5, "np.sin({})", r"\sin {}")
            >>> library.add_symbol("C", "const", 5, "C[{}]", r"c_{}")
            >>> library.add_symbol("X_0", "var", 5, "X[:, 0]", r"X_0")
            >>> library.add_symbol("pi", "lit", 5, "np.pi", r"\pi")

        Args:
            symbol: The symbol to be added to the library.
            symbol_type: The type of the symbol, one of "op" (operator), "fn" (function), "lit" (literal), "const" (constant), or "var" (variable).
            precedence: The precedence of the symbol, used to determine the order of operations.
            np_fn: A string representing the numpy function associated with this symbol.
            latex_str: A string that represents how the symbol is written in LaTeX
        """
        if latex_str is None:
            if symbol_type == "var":
                latex_str = f"{{}} \text{{{symbol}}} {{}}"
            elif symbol_type == "fn":
                latex_str = f"\text{{{symbol}}} {{}}"
            else:
                latex_str = f"\text{{{symbol}}}"

        if symbol_type == "var" and (np_fn is None or np_fn == ""):
            np_fn = "X[:, {}]".format(self.num_variables)

        if symbol_type == "var":
            self.num_variables += 1

        self.symbols[symbol] = {
            "symbol": symbol,
            "type": symbol_type,
            "precedence": precedence,
            "np_fn": np_fn,
            "latex_str": latex_str
        }

    def remove_symbol(self, symbol: str):
        """
        Removes a symbol from the library.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> len(library.symbols)
            1
            >>> library.remove_symbol("x")
            >>> len(library.symbols)
            0

        Args:
            symbol: The symbol to be removed from the library.

        Raises:
            KeyError: If the symbol does not exist in the library.
        """
        del self.symbols[symbol]

    def get_type(self, symbol: str) -> str:
        """
        Retrieves the type of a symbol from the library.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_type("x")
            'var'

        Args:
            symbol: The symbol whose type is to be retrieved.

        Returns:
            The type of the symbol if it exists in the library, otherwise an empty string.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["type"]
        else:
            return ""

    def get_precedence(self, symbol: str) -> int:
        """
        Retrieves the precedence of the given symbol.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_precedence("x")
            0

        Args:
            symbol: The symbol whose precedence is to be retrieved.

        Returns:
            The precedence of the symbol if it exists in the library, otherwise -1.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["precedence"]
        else:
            return -1

    def get_np_fn(self, symbol: str) -> str:
        """
        Returns the numpy function corresponding to the given symbol.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_np_fn("x")
            'x'

        Args:
            symbol: The symbol to look up.

        Returns:
            The numpy function corresponding to the given symbol, or an empty string if the symbol was not found.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["np_fn"]
        else:
            return ""

    def get_latex_str(self, symbol):
        """
        Returns the LaTeX template for the corresponding symbol.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x", "test")
            >>> library.get_latex_str("x")
            'test'

        Args:
            symbol: The symbol to look up.

        Returns:
            The LaTeX template for the corresponding symbol, or an empty string if the symbol was not found.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["latex_str"]
        else:
            return ""

    def get_symbols_of_type(self, symbol_type: str) -> List[str]:
        """
        Returns a list of symbols with the requested type ("op", "fn", "var", "const", "lit").

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.add_symbol("y", "var", 0, "y")
            >>> library.get_symbols_of_type("var")
            ['x', 'y']

        Args:
            symbol_type: Type of symbols you want to get.

        Returns:
            A list of symbols with the requested type
        """
        symbols = list()
        for symbol in self.symbols.keys():
            if self.get_type(symbol) == symbol_type:
                symbols.append(symbol)

        return symbols

    def symbols2index(self) -> Dict[str, int]:
        """
        Generates a dictionary mapping symbols to their indices in the symbol list.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.add_symbol("y", "var", 0, "y")
            >>> print(library.symbols2index())
            {'x': 0, 'y': 1}
            >>> library.remove_symbol("x")
            >>> print(library.symbols2index())
            {'y': 0}

        Returns:
            A dictionary mapping symbols to their indices in the symbol list.
        """
        return {s:i for i, s in enumerate(self.symbols.keys())}

    @staticmethod
    def from_symbol_list(symbols: List[str], num_variables=25):
        """
        Creates an instance of SymbolLibrary from a list of symbols and number of variables. The list of currently
        supported symbols (by default) can be seen in the SymbolLibrary.default_symbols() function.

        Examples:
            >>> library = SymbolLibrary().from_symbol_list(["+", "*", "C"], num_variables=2)
            >>> len(library.symbols)
            5

        Args:
            symbols: List of symbols you want.
            num_variables: Number of variables you want.

        Returns:
            An instance of SymbolLibrary
        """
        variables = [f"X_{i}" for i in range(num_variables)]
        symbols = symbols + variables

        sl = SymbolLibrary.default_symbols(num_variables)

        all_symbols = list(sl.symbols.keys())
        for symbol in all_symbols:
            if symbol not in symbols:
                sl.remove_symbol(symbol)

        return sl

    @staticmethod
    def default_symbols(num_variables: int = 25) -> "SymbolLibrary":
        """
        Creates a SymbolLibrary instance populated with default mathematical symbols.

        This method adds a set of predefined symbols to a SymbolLibrary instance,
        representing common mathematical operations, functions, constants, and optional
        variables. The symbols include basic arithmetic operations, trigonometric and
        exponential functions, and mathematical constants like pi and e.

        If num_variables is greater than 0, it adds variables labeled 'X_0' to 'X_{num_variables-1}', each
         associated with a column in a data array X.

        By default, we currently support the following symbols: "+", "-", "*", "/", "^", "u-" (unary minus), "sqrt",
        "sin", "cos", "exp", "tan", "arcsin", "arccos", "arctan", "sinh", "cosh", "tanh", "floor", "ceil", "ln", "log",
        "^-1", "^2", "^3", "^4", "^5", "pi", "e", "C" (unknown constant).

        Notes: The variables in the default_symbols function are added in the predefined order,
        which is the same order as the columns in the data array X.

        Examples:
            >>> library = SymbolLibrary.default_symbols()
            >>> len(library.symbols)
            54

        Args:
            num_variables: The number of variables to add to the library (default is 25).

        Returns:
            A SymbolLibrary instance populated with default mathematical symbols.
        """
        sl = SymbolLibrary()
        sl.add_symbol("+", symbol_type="op", precedence=0, np_fn="{} = {} + {}", latex_str=r"{} + {}")
        sl.add_symbol("-", symbol_type="op", precedence=0, np_fn="{} = {} - {}", latex_str=r"{} - {}")
        sl.add_symbol("*", symbol_type="op", precedence=1, np_fn="{} = {} * {}", latex_str=r"{} \cdot {}")
        sl.add_symbol("/", symbol_type="op", precedence=1, np_fn="{} = {} / {}", latex_str=r"\frac{{{}}}{{{}}}")
        sl.add_symbol("^", symbol_type="op", precedence=2, np_fn="{} = np.power({},{})", latex_str=r"{}^{{{}}}")
        sl.add_symbol("u-", symbol_type="fn", precedence=5, np_fn="{} = -{}", latex_str=r"- {}")
        sl.add_symbol("sqrt", symbol_type="fn", precedence=5, np_fn="{} = np.sqrt({})", latex_str=r"\sqrt {{{}}}")
        sl.add_symbol("sin", symbol_type="fn", precedence=5, np_fn="{} = np.sin({})", latex_str=r"\sin {}")
        sl.add_symbol("cos", symbol_type="fn", precedence=5, np_fn="{} = np.cos({})", latex_str=r"\cos {}")
        sl.add_symbol("exp", symbol_type="fn", precedence=5, np_fn="{} = np.exp({})", latex_str=r"e^{{{}}}")
        sl.add_symbol("tan", symbol_type="fn", precedence=5, np_fn="{} = np.tan({})", latex_str=r"\tan {}")
        sl.add_symbol("arcsin", symbol_type="fn", precedence=5, np_fn="{} = np.arcsin({})", latex_str=r"\arcsin {}")
        sl.add_symbol("arccos", symbol_type="fn", precedence=5, np_fn="{} = np.arccos({})", latex_str=r"\arccos {}")
        sl.add_symbol("arctan", symbol_type="fn", precedence=5, np_fn="{} = np.arctan({})", latex_str=r"\arctan {}")
        sl.add_symbol("sinh", symbol_type="fn", precedence=5, np_fn="{} = np.sinh({})", latex_str=r"\sinh {}")
        sl.add_symbol("cosh", symbol_type="fn", precedence=5, np_fn="{} = np.cosh({})", latex_str=r"\cosh {}")
        sl.add_symbol("tanh", symbol_type="fn", precedence=5, np_fn="{} = np.tanh({})", latex_str=r"\tanh {}")
        sl.add_symbol("floor", symbol_type="fn", precedence=5, np_fn="{} = np.floor({})", latex_str=r"\lfloor {} \rfloor")
        sl.add_symbol("ceil", symbol_type="fn", precedence=5, np_fn="{} = np.ceil({})", latex_str=r"\lceil {} \rceil")
        sl.add_symbol("ln", symbol_type="fn", precedence=5, np_fn="{} = np.log({})", latex_str=r"\ln {}")
        sl.add_symbol("log", symbol_type="fn", precedence=5, np_fn="{} = np.log10({})", latex_str=r"\log_{{10}} {}")
        sl.add_symbol("^-1", symbol_type="fn", precedence=-1, np_fn="{} = 1/{}", latex_str=r"{}^{{-1}}")
        sl.add_symbol("^2", symbol_type="fn", precedence=-1, np_fn="{} = {}**2", latex_str=r"{}^2")
        sl.add_symbol("^3", symbol_type="fn", precedence=-1, np_fn="{} = {}**3", latex_str=r"{}^3")
        sl.add_symbol("^4", symbol_type="fn", precedence=-1, np_fn="{} = {}**4", latex_str=r"{}^4")
        sl.add_symbol("^5", symbol_type="fn", precedence=-1, np_fn="{} = {}**5", latex_str=r"{}^5")
        sl.add_symbol("pi", symbol_type="lit", precedence=5, np_fn="np.full(X.shape[0], np.pi)", latex_str=r"\pi")
        sl.add_symbol("e", symbol_type="lit", precedence=5, np_fn="np.full(X.shape[0], np.e)", latex_str=r"e")
        sl.add_symbol("C", symbol_type="const", precedence=5, np_fn="np.full(X.shape[0], C[{}])", latex_str=r"C_{{{}}}")

        if num_variables > 0:
            for i in range(num_variables):
                sl.add_symbol(f"X_{i}", "var", 5, "X[:, {}]".format(i), "X_{{{}}}".format(i))

        return sl


"""
This module contains helper functions for creating a PCFG with generic probabilities from the SymbolLibrary and to use it for generating random expressions.
"""
from typing import Union, List

import nltk
import numpy as np
from tqdm import tqdm

from SRToolkit.utils import SymbolLibrary


def create_generic_pcfg(symbol_library: SymbolLibrary) -> str:
    """
    Creates a generic PCFG from the SymbolLibrary.

    Examples:
        >>> sl = SymbolLibrary.from_symbol_list(["+", "-", "*", "sin", "^2", "pi"], 2)
        >>> print(create_generic_pcfg(sl))
        E -> E '+' F [0.2]
        E -> E '-' F [0.2]
        E -> F [0.6]
        F -> F '*' B [0.4]
        F -> B [0.6]
        B -> T [1.0]
        T -> R [0.2]
        T -> C [0.2]
        T -> V [0.6]
        C -> 'pi' [1.0]
        R -> 'sin' '(' E ')' [0.4]
        R -> P [0.15]
        R -> '(' E ')' [0.45]
        P -> '(' E ')' '^2' [1.0]
        V -> 'X_0' [0.5]
        V -> 'X_1' [0.5]
        <BLANKLINE>

    Args:
        symbol_library: The symbol library to use. Defaults to SymbolLibrary.default_symbols().

    Returns:
        A PCFG with generic probabilities, written as a string.
    """
    symbols = symbol_library.symbols.values()
    E = [s["symbol"] for s in symbols if s["type"]=="op" and s["precedence"]==0]
    F = [s["symbol"] for s in symbols if s["type"]=="op" and s["precedence"]==1]
    BP = [s["symbol"] for s in symbols if s["type"]=="op" and s["precedence"]==2]
    R = [s["symbol"] for s in symbols if s["type"]=="fn" and s["precedence"]==5]
    P = [s["symbol"] for s in symbols if s["type"]=="fn" and s["precedence"]==-1]
    V = [s["symbol"] for s in symbols if s["type"]=="var"]
    Cc = [s["symbol"] for s in symbols if s["type"]=="const"]
    Cl = [s["symbol"] for s in symbols if s["type"]=="lit"]

    grammar = ""
    if len(E) > 0:
        for s in E:
            grammar += f"E -> E '{s}' F [{0.4/len(E)}]\n"
        grammar += "E -> F [0.6]\n"
    else:
        grammar += "E -> F [1.0]\n"

    if len(F) > 0:
        for s in F:
            grammar += f"F -> F '{s}' B [{0.4/len(F)}]\n"
        grammar += "F -> B [0.6]\n"
    else:
        grammar += "F -> B [1.0]\n"

    if len(BP) > 0:
        for s in BP:
            grammar += f"B -> B '{s}' T [{0.05/len(BP)}]\n"
        grammar += "B -> T [0.95]\n"
    else:
        grammar += "B -> T [1.0]\n"

    if len(Cc) + len(Cl) > 0:
        grammar += "T -> R [0.2]\n"
        grammar += "T -> C [0.2]\n"
        grammar += "T -> V [0.6]\n"
        if len(Cl) > 0 and len(Cc) > 0:
            for s in Cl:
                grammar += f"C -> '{s}' [{0.2/len(Cl)}]\n"
            for s in Cc:
                grammar += f"C -> '{s}' [{0.8/len(Cc)}]\n"
        elif len(Cl) > 0:
            for s in Cl:
                grammar += f"C -> '{s}' [{1/len(Cl)}]\n"
        elif len(Cc) > 0:
            for s in Cc:
                grammar += f"C -> '{s}' [{1/len(Cc)}]\n"
    else:
        grammar += "T -> R [0.3]\n"
        grammar += "T -> V [0.7]\n"

    if len(R) > 0:
        for s in R:
            grammar += f"R -> '{s}' '(' E ')' [{0.4/len(R)}]\n"
        if len(P) > 0:
            grammar += "R -> P [0.15]\n"
            grammar += "R -> '(' E ')' [0.45]\n"
        else:
            grammar += "R -> '(' E ')' [0.6]\n"
    else:
        if len(P) > 0:
            grammar += "R -> P [0.15]\n"
            grammar += "R -> '(' E ')' [0.85]\n"
        else:
            grammar += "R -> '(' E ')' [1.0]\n"

    if len(P) > 0:
        total = sum([1/abs(float(s[1:])) for s in P])
        for s in P:
            grammar += f"P -> '(' E ')' '{s}' [{(1/abs(float(s[1:])))/total}]\n"

    if len(V) > 0:
        for s in V:
            grammar += f"V -> '{s}' [{1/len(V)}]\n"

    return grammar


def _expand(grammar, symbol, current_depth, max_depth=40):
    if current_depth > max_depth > 0:
        return None

    if isinstance(symbol, nltk.grammar.Nonterminal):
        productions = grammar.productions(lhs=symbol)
        if not productions:
            return [str(symbol)]

        weights = [p.prob() for p in productions]
        chosen_production = np.random.choice(productions, p=weights)

        generated_sequence = []
        for rhs_symbol in chosen_production.rhs():
            expanded_part = _expand(grammar, rhs_symbol, current_depth + 1)
            if expanded_part is None:
                return None
            generated_sequence.extend(expanded_part)
        return generated_sequence
    else:
        return [str(symbol)]


def generate_from_pcfg(grammar_str: str, start_symbol="E", max_depth=40, limit=100) -> List[str]:
    """
    Generates a random expression from a PCFG with monte-carlo sampling.

    Examples:
        >>> generate_from_pcfg("E -> '1' [1.0]")
        ['1']
        >>> grammar = create_generic_pcfg(SymbolLibrary.default_symbols())
        >>> len(generate_from_pcfg(grammar)) > 0
        True

    Args:
        grammar_str: Grammar given as a string in the NLTK notation
        start_symbol: Non-terminal symbol used as the starting point
        max_depth: Maximum depth of the generated parse trees. If less than 0, expressions can have arbitrary depth
        limit: Number of times the function tries to generate a valid expression before raising an Exception.

    Raises:
        Exception: If the maximum number of tries is reached without generating a valid expression

    Returns:
        An expression written as a list of string tokens in the infix notation.
    """
    start_symbol = nltk.grammar.Nonterminal(start_symbol)
    grammar = nltk.PCFG.fromstring(grammar_str)
    expr = _expand(grammar, start_symbol, 0, max_depth)
    tries = 1
    while expr is None and tries < limit:
        expr = _expand(grammar, start_symbol, 0, max_depth)

    if expr is None:
        raise Exception(f"[Expression generation] Couldn't find an expression with max_depth {max_depth} from this grammar in {limit} tries.")

    return expr


def generate_n_expressions(expression_description: Union[str, SymbolLibrary], num_expressions: int, unique=True,
                            max_expression_length=50, verbose=True) -> List[List[str]]:
    """
    Generates a set of n expressions.

    Examples:
        >>> len(generate_n_expressions(SymbolLibrary.default_symbols(5), 100, verbose=False))
        100
        >>> generate_n_expressions(SymbolLibrary.from_symbol_list([], 1), 3, unique=False, verbose=False, max_expression_length=1)
        [['X_0'], ['X_0'], ['X_0']]

    Args:
        expression_description: Decription of expressions, given as either a grammar in the NLTK notation or a SymbolLibrary instance
        num_expressions: Number of generated expressions
        unique: When True, each generated expression will be unique (not necesarily unequivalent to others)
        max_expression_length: Generated expressions will have at most "max_expression_length" tokens. If less than 0, expressions can be of arbitrary size.
        verbose: If True, adds a progress bar

    Returns:
        A list of expressions represented as lists of tokens
    """
    if isinstance(expression_description, SymbolLibrary):
        grammar = create_generic_pcfg(expression_description)
    elif isinstance(expression_description, str):
        grammar = expression_description
    else:
        raise Exception("Description of expressions must be either a grammar written as a string or an instance of SymbolLibrary.")

    expressions = []
    expression_strings = set()
    if verbose:
        pbar = tqdm(total=num_expressions)
    while len(expressions) < num_expressions:
        try:
            expr = generate_from_pcfg(grammar, max_depth=max_expression_length*10)
        except:
            print("Couldn't generate a valid expression in 100 tries")
            continue
        if len(expr) > max_expression_length > 0:
            continue

        expr_string = "".join(expr)
        if expr_string not in expression_strings or not unique:
            expressions.append(expr)
            expression_strings.add(expr_string)
            if verbose:
                pbar.update(1)

    if verbose:
        pbar.close()
    return expressions


if __name__ == '__main__':
    sl = SymbolLibrary.default_symbols(5)
    a = generate_n_expressions(sl, 1000, unique=False, max_expression_length=1)
    b = 0
    # grammar = create_generic_pcfg(sl)
    # print(generate_from_pcfg(grammar))
